package io.tswf.groovy.bettergroovy.asm.member.filters.standard

import com.google.auto.service.AutoService
import groovy.transform.CompileStatic
import io.tswf.groovy.bettergroovy.asm.member.FieldNodeMember
import io.tswf.groovy.bettergroovy.asm.member.IMember
import io.tswf.groovy.bettergroovy.asm.member.MethodNodeMember
import io.tswf.groovy.bettergroovy.asm.member.filters.IGroovyInternalMemberFilter
import org.codehaus.groovy.reflection.ClassInfo
import org.objectweb.asm.Type

import java.util.concurrent.ConcurrentHashMap
import java.util.concurrent.CopyOnWriteArrayList

@AutoService(IGroovyInternalMemberFilter)
@CompileStatic
class GroovyNotAnnotatedInternalMemberFilter implements IGroovyInternalMemberFilter {

    public static final Type CLASS_INFO_TYPE = ClassInfo.asAsmType()
    public static final Type GROOVY_OBJECT_TYPE = GroovyObject.asAsmType()
    public static final Type GROOVY_META_CLASS_TYPE = MetaClass.asAsmType()
    public static final Type METHOD_HANDLES_TYPE = Type.getObjectType("java/lang/invoke/MethodHandles")
    public static final Type METHOD_HANDLES_LOOKUP_TYPE = Type.getObjectType('java/lang/invoke/MethodHandles$Lookup')

    public static final String GET_STATIC_META_CLASS_METHOD_NAME = '$getStaticMetaClass'

    public static final String GET_LOOKUP_METHOD_NAME = '$getLookup'

    public static final String STATIC_CLASS_INFO_FIELD_NAME = '$staticClassInfo$'
    public static final String STATIC_CLASS_INFO_FIELD_NAME_2 = '$staticClassInfo'

    public static final String META_CLASS_FIELD_NAME = 'metaClass'

    public static final String STMC_FIELD_NAME = '__$stMC'

    Map<String, Set<Type>> internalTypes

    GroovyNotAnnotatedInternalMemberFilter() {
        internalTypes = new ConcurrentHashMap<>()

        registerInternalMember(GET_STATIC_META_CLASS_METHOD_NAME, GROOVY_META_CLASS_TYPE)
        registerInternalMember(GET_LOOKUP_METHOD_NAME, METHOD_HANDLES_LOOKUP_TYPE)

        registerInternalMember(STATIC_CLASS_INFO_FIELD_NAME, CLASS_INFO_TYPE)
        registerInternalMember(STATIC_CLASS_INFO_FIELD_NAME_2, CLASS_INFO_TYPE)
        registerInternalMember(META_CLASS_FIELD_NAME, GROOVY_META_CLASS_TYPE)
        registerInternalMember(STMC_FIELD_NAME, Type.BOOLEAN_TYPE)
    }

    void registerInternalMember(String memberName, Type memberType) {
        getTypesByName(memberName, true) << memberType
    }

    Set<Type> getTypesByName(String memberName, boolean create = false) {
        if (create) {
            return internalTypes.computeIfAbsent(memberName) {
                Collections.concurrentSet()
            }
        } else {
            return internalTypes[memberName]
        }
    }

    @Override
    boolean isGroovyInternal(IMember member) {
        return getTypesByName(member.name)?.contains(member.type) ?: false
    }
}
