package io.tswf.groovy.bettergroovy.asm.textifier

import groovy.transform.CompileStatic
import org.objectweb.asm.tree.ClassNode
import org.objectweb.asm.tree.MethodNode
import org.objectweb.asm.util.Textifier
import org.objectweb.asm.util.TraceClassVisitor
import org.objectweb.asm.util.TraceMethodVisitor

import java.util.function.Consumer

@CompileStatic
class TextifyUtils {

    static String textifyMethod(MethodNode methodNode) {
        textify {
            methodNode.accept(new TraceMethodVisitor(null, it))
        }
    }

    static String textifyClass(ClassNode classNode) {
        def stringWriter = new StringWriter()
        def printWriter = new PrintWriter(stringWriter)

        classNode.accept(new TraceClassVisitor(printWriter))

        return stringWriter.toString()
    }

    static String textify(Consumer<Textifier> textifierConsumer) {
        def textifier = new Textifier()
        textifierConsumer.accept(textifier)
        return textifier.text.join("")
    }
}
